package dev.fyloz.colorrecipesexplorer.logic

import dev.fyloz.colorrecipesexplorer.Constants
import dev.fyloz.colorrecipesexplorer.config.annotations.LogicComponent
import dev.fyloz.colorrecipesexplorer.dtos.*
import dev.fyloz.colorrecipesexplorer.exception.AlreadyExistsException
import dev.fyloz.colorrecipesexplorer.logic.files.WriteableFileLogic
import dev.fyloz.colorrecipesexplorer.logic.users.GroupLogic
import dev.fyloz.colorrecipesexplorer.model.Recipe
import dev.fyloz.colorrecipesexplorer.service.RecipeService
import dev.fyloz.colorrecipesexplorer.utils.collections.LazyMapList
import dev.fyloz.colorrecipesexplorer.utils.merge
import org.springframework.transaction.annotation.Transactional
import org.springframework.web.multipart.MultipartFile

interface RecipeLogic : Logic<RecipeDto, RecipeService> {
    /** Gets all recipes and load their mixes and groupsInformation, to prevent LazyInitializationExceptions */
    fun getAllWithMixesAndGroupsInformation(): Collection<RecipeDto>

    /** Gets all recipes with the given [name]. */
    fun getAllByName(name: String): Collection<RecipeDto>

    /** Saves the given [dto]. */
    fun save(dto: RecipeSaveDto): RecipeDto

    /** Updates the given [dto]. */
    fun update(dto: RecipeUpdateDto): RecipeDto

    /** Updates the public data of a recipe with the given [publicDataDto]. */
    fun updatePublicData(publicDataDto: RecipePublicDataDto)
}

@LogicComponent
class DefaultRecipeLogic(
    service: RecipeService,
    private val companyLogic: CompanyLogic,
    private val recipeStepLogic: RecipeStepLogic,
    private val mixLogic: MixLogic,
    private val groupLogic: GroupLogic
) : BaseLogic<RecipeDto, RecipeService>(service, Recipe::class.simpleName!!), RecipeLogic {
    @Transactional
    override fun getAllWithMixesAndGroupsInformation() =
        getAll().onEach { (it.mixes as LazyMapList<*, *>).initialize() }
            .onEach { (it.groupsInformation as LazyMapList<*, *>).initialize() }

    override fun getAllByName(name: String) = service.getAllByName(name)

    override fun save(dto: RecipeSaveDto) = save(
        RecipeDto(
            name = dto.name,
            description = dto.description,
            color = dto.color,
            gloss = dto.gloss,
            sample = dto.sample,
            approbationDate = dto.approbationDate,
            approbationExpired = false,
            remark = dto.remark ?: "",
            company = companyLogic.getById(dto.companyId),
            mixes = listOf(),
            groupsInformation = listOf()
        )
    )

    override fun save(dto: RecipeDto): RecipeDto {
        throwIfNameAndCompanyAlreadyExists(dto.name, dto.company.id)

        return super.save(dto)
    }

    override fun update(dto: RecipeUpdateDto): RecipeDto {
        val recipe = getById(dto.id)

        return update(
            RecipeDto(
                id = dto.id,
                name = dto.name,
                description = dto.description,
                color = dto.color,
                gloss = dto.gloss,
                sample = dto.sample,
                approbationDate = dto.approbationDate,
                approbationExpired = false,
                remark = dto.remark ?: "",
                company = recipe.company,
                mixes = recipe.mixes,
                groupsInformation = updateGroupsInformationSteps(recipe, dto)
            )
        )
    }

    override fun update(dto: RecipeDto): RecipeDto {
        throwIfNameAndCompanyAlreadyExists(dto.name, dto.company.id, dto.id)

        return super.update(dto)
    }

    @Transactional
    override fun updatePublicData(publicDataDto: RecipePublicDataDto) {
        // Update notes
        if (publicDataDto.notes.isNotEmpty()) {
            val recipe = getById(publicDataDto.recipeId)
            update(recipe.copy(groupsInformation = updateGroupsInformationNotes(recipe, publicDataDto.notes)))
        }

        // Update mixes locations
        if (publicDataDto.mixesLocation.isNotEmpty()) {
            mixLogic.updateLocations(publicDataDto.mixesLocation)
        }
    }

    private fun updateGroupsInformationSteps(recipe: RecipeDto, dto: RecipeUpdateDto): List<RecipeGroupInformationDto> {
        val updatedGroupsInformation = dto.steps.map { updateGroupInformationSteps(recipe, it) }
        return recipe.groupsInformation.merge(updatedGroupsInformation)
    }

    private fun updateGroupInformationSteps(recipe: RecipeDto, groupSteps: RecipeGroupStepsDto) =
        getOrCreateGroupInformation(recipe, groupSteps.groupId).copy(steps = groupSteps.steps).also {
            recipeStepLogic.validateGroupInformationSteps(it)
        }

    private fun updateGroupsInformationNotes(
        recipe: RecipeDto, notes: List<RecipeGroupNoteDto>
    ): List<RecipeGroupInformationDto> {
        val updatedGroupsInformation = notes.map { updateGroupInformationNote(recipe, it) }
        return recipe.groupsInformation.merge(updatedGroupsInformation)
    }

    private fun updateGroupInformationNote(recipe: RecipeDto, groupNote: RecipeGroupNoteDto) =
        getOrCreateGroupInformation(recipe, groupNote.groupId).copy(note = groupNote.content)

    private fun getOrCreateGroupInformation(recipe: RecipeDto, groupId: Long) =
        recipe.groupsInformation.firstOrNull { it.group.id == groupId }
            ?: RecipeGroupInformationDto(group = groupLogic.getById(groupId))

    private fun throwIfNameAndCompanyAlreadyExists(name: String, companyId: Long, id: Long? = null) {
        if (service.existsByNameAndCompany(name, companyId, id)) {
            throw AlreadyExistsException(
                "$typeNameLowerCase-company",
                "$typeName already exists",
                "A recipe with the name '$name' already exists for the company with the id '$companyId'",
                name,
                NAME_IDENTIFIER_NAME,
                mutableMapOf(
                    "companyId" to companyId
                )
            )
        }
    }
}

interface RecipeImageLogic {
    /** Gets the id of every image associated to the recipe with the given [recipeId]. */
    fun getAllImages(recipeId: Long): List<String>

    /** Saves the given [image] and associate it to the recipe with the given [recipeId]. Returns the id of the saved image. */
    fun download(image: MultipartFile, recipeId: Long): String

    /** Deletes the image with the given [path] for the given [recipeId]. */
    fun delete(recipeId: Long, path: String)
}

@LogicComponent
class DefaultRecipeImageLogic(val fileLogic: WriteableFileLogic) : RecipeImageLogic {
    override fun getAllImages(recipeId: Long) =
        fileLogic.listDirectoryFiles(getRecipeImagesDirectory(recipeId)).map { it.name }

    override fun download(image: MultipartFile, recipeId: Long): String {
        /** Gets the next id available for a new image for the given [recipeId]. */
        fun getNextAvailableId(): String = with(getAllImages(recipeId)) {
            (if (isEmpty()) 0 else maxOf { it.toLong() } + 1L).toString()
        }

        return getNextAvailableId().also {
            val imagePath = getImagePath(recipeId, it)
            fileLogic.writeToDirectory(image, imagePath, getRecipeImagesDirectory(recipeId), true)
        }
    }

    override fun delete(recipeId: Long, path: String) =
        fileLogic.deleteFromDirectory(path, getRecipeImagesDirectory(recipeId))

    private fun getImagePath(recipeId: Long, id: String) = "${getRecipeImagesDirectory(recipeId)}/$id"

    private fun getRecipeImagesDirectory(recipeId: Long) = "${Constants.FilePaths.RECIPE_IMAGES}/$recipeId"
}
