package dev.fyloz.colorrecipesexplorer.logic

import dev.fyloz.colorrecipesexplorer.Constants
import dev.fyloz.colorrecipesexplorer.logic.files.CachedFile
import dev.fyloz.colorrecipesexplorer.logic.files.WriteableFileLogic
import dev.fyloz.colorrecipesexplorer.utils.FilePath
import io.mockk.*
import org.junit.jupiter.api.AfterEach
import org.junit.jupiter.api.Test
import org.springframework.mock.web.MockMultipartFile
import kotlin.test.assertEquals

class DefaultRecipeImageLogicTest {
    private val fileLogicMock = mockk<WriteableFileLogic>()

    private val recipeImageLogic = spyk(DefaultRecipeImageLogic(fileLogicMock))

    private val recipeId = 1L

    @AfterEach
    internal fun afterEach() {
        clearAllMocks()
    }

    @Test
    fun getAllImages_normalBehavior_returnsAllRecipeImages() {
        // Arrange
        val filePath = FilePath("${Constants.FilePaths.RECIPE_IMAGES}/$recipeId")
        val files = listOf(
            CachedFile("0", filePath, true),
            CachedFile("1", filePath, true)
        )
        val expectedImages = files.map { it.name }

        every { fileLogicMock.listDirectoryFiles(any()) } returns files

        // Act
        val actualImages = recipeImageLogic.getAllImages(recipeId)

        // Assert
        assertEquals(expectedImages, actualImages)
    }

    @Test
    fun download_normalBehavior_callsWriteToDirectoryInFileLogic() {
        // Arrange
        val previousImageId = 0L

        every { recipeImageLogic.getAllImages(recipeId) } returns listOf(previousImageId.toString())
        every { fileLogicMock.writeToDirectory(any(), any(), any(), any()) } just runs

        val file = MockMultipartFile("Unit test name", byteArrayOf())

        val expectedFilePath = "${Constants.FilePaths.RECIPE_IMAGES}/$recipeId"
        val expectedImageId = previousImageId + 1

        // Act
        recipeImageLogic.download(file, recipeId)

        // Assert
        verify {
            fileLogicMock.writeToDirectory(file, "$expectedFilePath/$expectedImageId", expectedFilePath, true)
        }
    }

    @Test
    fun download_normalBehavior_returnsImageId() {
        // Arrange
        val previousImageId = 0L

        every { recipeImageLogic.getAllImages(recipeId) } returns listOf(previousImageId.toString())
        every { fileLogicMock.writeToDirectory(any(), any(), any(), any()) } just runs

        val file = MockMultipartFile("Unit test name", byteArrayOf())

        val expectedImageId = previousImageId + 1

        // Act
        val downloadedImageId = recipeImageLogic.download(file, recipeId)

        // Assert
        assertEquals(expectedImageId.toString(), downloadedImageId)
    }

    @Test
    fun delete_normalBehavior_callsDeleteFromDirectoryInFileLogic() {
        // Arrange
        every { fileLogicMock.deleteFromDirectory(any(), any()) } just runs

        val recipeImagesDirectoryPath = "${Constants.FilePaths.RECIPE_IMAGES}/$recipeId"
        val imagePath = "$recipeImagesDirectoryPath/1"

        // Act
        recipeImageLogic.delete(recipeId, imagePath)

        // Assert
        verify {
            fileLogicMock.deleteFromDirectory(imagePath, recipeImagesDirectoryPath)
        }
    }
}