package dev.fyloz.colorrecipesexplorer.logic

import dev.fyloz.colorrecipesexplorer.dtos.*
import dev.fyloz.colorrecipesexplorer.exception.AlreadyExistsException
import dev.fyloz.colorrecipesexplorer.logic.users.GroupLogic
import dev.fyloz.colorrecipesexplorer.model.account.Group
import dev.fyloz.colorrecipesexplorer.service.RecipeService
import io.mockk.*
import org.junit.jupiter.api.AfterEach
import org.junit.jupiter.api.Test
import org.junit.jupiter.api.assertThrows
import kotlin.test.assertContains
import kotlin.test.assertEquals

class DefaultRecipeLogicTest {
    private val recipeServiceMock = mockk<RecipeService>()
    private val companyLogicMock = mockk<CompanyLogic>()
    private val recipeStepLogicMock = mockk<RecipeStepLogic>()
    private val mixLogicMock = mockk<MixLogic>()
    private val groupLogicMock = mockk<GroupLogic>()

    private val recipeLogic =
        spyk(DefaultRecipeLogic(recipeServiceMock, companyLogicMock, recipeStepLogicMock, mixLogicMock, groupLogicMock))

    private val company = CompanyDto(1L, "Unit test company")
    private val group = Group(1L, "Unit test group")
    private val recipe = RecipeDto(
        1L,
        "Unit test recipe",
        "Unit test recipe",
        "FFFFFF",
        0xf,
        null,
        null,
        false,
        "Remark",
        company,
        listOf(),
        listOf()
    )

    @AfterEach
    internal fun afterEach() {
        clearAllMocks()
    }

    @Test
    fun getAllByName_normalBehavior_returnsFromService() {
        // Arrange
        val expectedRecipes = listOf(recipe)

        every { recipeServiceMock.getAllByName(any()) } returns expectedRecipes

        // Act
        val actualRecipes = recipeLogic.getAllByName(recipe.name)

        // Assert
        assertEquals(actualRecipes, expectedRecipes)
    }

    @Test
    fun save_dto_normalBehavior_returnsFromSave() {
        // Arrange
        every { recipeServiceMock.existsByNameAndCompany(any(), any(), any()) } returns false
        every { companyLogicMock.getById(any()) } returns company
        every { recipeLogic.save(any<RecipeDto>()) } returns recipe

        val dto = RecipeSaveDto(
            recipe.name,
            recipe.description,
            recipe.color,
            recipe.gloss,
            recipe.sample,
            recipe.approbationDate,
            recipe.remark,
            company.id
        )

        // Act
        val savedRecipe = recipeLogic.save(dto)

        // Assert
        assertEquals(recipe, savedRecipe)
    }

    @Test
    fun save_nameAndCompanyExists_throwsAlreadyExistsException() {
        // Arrange
        every { recipeServiceMock.existsByNameAndCompany(any(), any(), any()) } returns true

        // Act
        // Assert
        assertThrows<AlreadyExistsException> { recipeLogic.save(recipe) }
    }

    @Test
    fun update_dto_normalBehavior_returnsFromSave() {
        // Arrange
        every { recipeServiceMock.existsByNameAndCompany(any(), any(), any()) } returns false
        every { recipeServiceMock.getById(any()) } returns recipe
        every { companyLogicMock.getById(any()) } returns company
        every { recipeLogic.update(any<RecipeDto>()) } returns recipe

        val dto = RecipeUpdateDto(
            recipe.id,
            recipe.name,
            recipe.description,
            recipe.color,
            recipe.gloss,
            recipe.sample,
            recipe.approbationDate,
            recipe.remark,
            listOf()
        )

        // Act
        val updatedRecipe = recipeLogic.update(dto)

        // Assert
        assertEquals(recipe, updatedRecipe)
    }

    @Test
    fun update_nameAndCompanyExists_throwsAlreadyExistsException() {
        // Arrange
        every { recipeServiceMock.existsByNameAndCompany(any(), any(), any()) } returns true

        // Act
        // Assert
        assertThrows<AlreadyExistsException> { recipeLogic.update(recipe) }
    }

    @Test
    fun updatePublicData_normalBehavior_callsUpdate() {
        // Arrange
        every { recipeLogic.getById(any()) } returns recipe
        every { recipeLogic.update(any<RecipeDto>()) } returnsArgument 0
        every { groupLogicMock.getById(any()) } returns group

        val groupNote = RecipeGroupNoteDto(1L, "Unit test note")
        val dto = RecipePublicDataDto(recipe.id, listOf(groupNote), listOf())

        // Act
        recipeLogic.updatePublicData(dto)

        // Assert
        verify {
            recipeLogic.update(any<RecipeDto>())
        }
    }

    @Test
    fun updatePublicData_normalBehavior_updatesRecipeGroupsInformation() {
        // Arrange
        var updatedRecipe = recipe

        every { recipeLogic.getById(any()) } returns recipe
        every { recipeLogic.update(any<RecipeDto>()) } answers { firstArg<RecipeDto>().also { updatedRecipe = it } }
        every { groupLogicMock.getById(any()) } returns group

        val expectedGroupInformation = RecipeGroupInformationDto(0L, group, "Unit test note", listOf())

        val groupNote = RecipeGroupNoteDto(group.id!!, expectedGroupInformation.note)
        val dto = RecipePublicDataDto(recipe.id, listOf(groupNote), listOf())

        // Act
        recipeLogic.updatePublicData(dto)

        // Assert
        assertContains(updatedRecipe.groupsInformation, expectedGroupInformation)
    }

    @Test
    fun updatePublicData_emptyNotes_doesNothing() {
        // Arrange
        val dto = RecipePublicDataDto(recipe.id, listOf(), listOf())

        // Act
        recipeLogic.updatePublicData(dto)

        // Assert
        verify(exactly = 0) {
            recipeLogic.update(any<RecipeDto>())
        }
    }

    @Test
    fun updatePublicData_normalBehavior_callsUpdateLocationsInMixLogic() {
        // Arrange
        every { mixLogicMock.updateLocations(any()) } just runs

        val mixesLocation = listOf(MixLocationDto(group.id!!, "location"))
        val dto = RecipePublicDataDto(recipe.id, listOf(), mixesLocation)

        // Act
        recipeLogic.updatePublicData(dto)

        // Assert
        verify {
            mixLogicMock.updateLocations(mixesLocation)
        }
    }

    @Test
    fun updatePublicData_emptyMixesLocation_doesNothing() {
        // Arrange
        val dto = RecipePublicDataDto(recipe.id, listOf(), listOf())

        // Act
        recipeLogic.updatePublicData(dto)

        // Assert
        verify(exactly = 0) {
            mixLogicMock.updateLocations(any())
        }
    }
}